"""A daemon exposing D-Bus services for managing the charge level of recent
ASUS notebooks.
"""

import logging
import sys
from os import path

__version__ = "0.3.1"

APP_NAME = "asuscharged"
FRIENDLY_NAME = "ASUS Battery Charge Control Daemon"
PACKAGE_NAME = "asus-charge-daemon"
STATE_DIR = "/var/lib/asuscharged"
STATE_FILE = "charge_control_end_threshold"
STATE_PATH = path.join(STATE_DIR, STATE_FILE)
CONFIG_DIR = "/etc/asuscharged"
CONFIG_FILE = "asuscharged.conf"
CONFIG_PATH = path.join(CONFIG_DIR, CONFIG_FILE)
DEV_CONFIG_DIR = path.join(path.dirname(path.dirname(__file__)), "dev")
DEV_CONFIG_PATH = path.join(DEV_CONFIG_DIR, CONFIG_FILE)
DBUS_NAME = "ca.cforrester.AsusChargeDaemon1"
DBUS_PATH = "/ca/cforrester/AsusChargeDaemon1"
NOTIFICATION_ICON = "battery-charging-good"

if __debug__:
    log_level = logging.DEBUG
    log_format = "%(asctime)s [%(levelname)s] %(name)s(%(lineno)d) - %(message)s"
else:
    log_level = logging.WARNING
    log_format = "[%(levelname)s] %(message)s"

logging.basicConfig(stream=sys.stdout, level=log_level, format=log_format)



#!/usr/bin/env python3
import os
from os import path
import shutil
import stat
import subprocess
import sys
from typing import Dict, List

APP_NAME = "asuscharged"
PACKAGE_NAME = "asus-charge-daemon"
DATADIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
SYSTEMD_SYSTEM_UNIT_FILE = "asuscharged.service"
DBUS_INTERFACE_FILE = "ca.cforrester.AsusChargeDaemon1.xml"
DBUS_SYSTEM_SERVICE_FILE = "ca.cforrester.AsusChargeDaemon1.service"
DBUS_SYSTEM_CONF_FILE = "ca.cforrester.AsusChargeDaemon1.conf"

files: Dict[str, str] = {}


def has_systemd() -> bool:
    return os.path.exists("/run/systemd/system")


def systemd_system_unit_dir() -> str:
    return subprocess.run(
        ("pkg-config", "systemd", "--variable=systemdsystemunitdir"),
        text=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()


def dbus_interfaces_dir() -> str:
    return subprocess.run(
        ("pkg-config", "dbus-1", "--variable=interfaces_dir"),
        text=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()


def dbus_system_services_dir() -> str:
    return subprocess.run(
        ("pkg-config", "dbus-1", "--variable=system_bus_services_dir"),
        text=True,
        stdout=subprocess.PIPE,
    ).stdout.strip()


def dbus_system_conf_dir() -> str:
    return path.join(
        subprocess.run(
            ("pkg-config", "dbus-1", "--variable=sysconfdir"),
            text=True,
            stdout=subprocess.PIPE,
        ).stdout.strip(),
        "dbus-1",
        "system.d",
    )


def files_exist() -> List[str]:
    exists = []
    for file in files.values():
        if os.path.exists(file):
            exists.append(file)
    return exists


def install(install_daemon=True, force=False, local=False) -> None:
    if not force:
        exists = files_exist()
        if exists:
            exists = "\n".join(exists)
            print(f"The following files exist:\n{exists}")
            if not input("Overwrite? [Y/N] ").startswith(("y", "Y")):
                raise SystemExit("Not overwriting.")
        for file in files:
            dest = files[file]
            print(f"Installing {file} to {os.path.dirname(dest)}...", end=" ")
            source = os.path.join(DATADIR, file)
            shutil.copy(source, dest)
            os.chmod(dest, stat.S_IREAD | stat.S_IWRITE | stat.S_IRGRP | stat.S_IROTH)
            print("\033[01m\033[32m✓\033[0m")
    if install_daemon:
        if local:
            print(f"Installing {APP_NAME} package from local directory...")
            subprocess.run((sys.executable, "-m", "pip", "install", os.curdir))
        else:
            print(f"Installing {APP_NAME} package from PyPI...")
            subprocess.run((sys.executable, "-m", "pip", "install", PACKAGE_NAME))
    print("Enabling and running systemd service...", end=" ")
    subprocess.run(("systemctl", "enable", APP_NAME), capture_output=True)
    subprocess.run(("systemctl", "start", APP_NAME), capture_output=True)
    print("\033[01m\033[32m✓\033[0m")


def uninstall(uninstall_daemon=True) -> None:
    print("Stopping and disabling systemd service...", end=" ")
    subprocess.run(("systemctl", "stop", APP_NAME), capture_output=True)
    subprocess.run(("systemctl", "disable", APP_NAME), capture_output=True)
    print("\033[01m\033[32m✓\033[0m")
    exists = files_exist()
    if not exists:
        print("No data files found, skipping...")
    else:
        for file in exists:
            print(f"Deleting {file}...", end=" ")
            os.remove(file)
            print("\033[01m\033[32m✓\033[0m")
    if uninstall_daemon:
        subprocess.run((sys.executable, "-m", "pip", "uninstall", "-y", PACKAGE_NAME))


if __name__ == "__main__":
    if os.geteuid() != 0:
        raise SystemExit("Script must be run as root.")
    if not has_systemd():
        raise SystemExit("Script is only compatible with systems running systemd.")
    try:
        files = {
            SYSTEMD_SYSTEM_UNIT_FILE: os.path.join(
                systemd_system_unit_dir(), SYSTEMD_SYSTEM_UNIT_FILE
            ),
            DBUS_INTERFACE_FILE: os.path.join(
                dbus_interfaces_dir(), DBUS_INTERFACE_FILE
            ),
            DBUS_SYSTEM_SERVICE_FILE: os.path.join(
                dbus_system_services_dir(), DBUS_SYSTEM_SERVICE_FILE
            ),
            DBUS_SYSTEM_CONF_FILE: os.path.join(
                dbus_system_conf_dir(), DBUS_SYSTEM_CONF_FILE
            ),
        }
        if sys.argv[1] == "install":
            try:
                if sys.argv[2] == "local":
                    install(local=True)
                else:
                    install()
            except IndexError:
                install()
        elif sys.argv[1] == "uninstall":
            uninstall()
        elif sys.argv[1] == "reinstall":
            uninstall()
            try:
                if sys.argv[2] == "local":
                    install(local=True)
                else:
                    install()
            except IndexError:
                install()
        else:
            raise SystemExit("Missing argument: install or uninstall.")
    except IndexError:
        raise SystemExit("Missing argument: install or uninstall.")



import configparser
import logging

from asuscharged import CONFIG_PATH, DEV_CONFIG_PATH

_log = logging.getLogger(__name__)

config = configparser.ConfigParser()
config["daemon"] = {
    "restore_on_start": "yes",
    "notify_on_restore": "yes",
    "notify_on_change": "yes",
}

if __debug__:
    _files = (CONFIG_PATH, DEV_CONFIG_PATH)
else:
    _files = CONFIG_PATH
_files_read = config.read(_files)
if _files_read:
    _log.debug(f"Loaded configuration from: {_files_read}")
else:
    _log.debug(
        f"No valid configuration files found. Exporting defaults to {CONFIG_PATH}"
    )
    with open(CONFIG_PATH, "w") as file:
        config.write(file)



# type: ignore

import logging
import subprocess
import os
import pwd
from dataclasses import dataclass
from enum import Enum
from os import path
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from typing import Dict, Optional, Set

import gi

gi.require_version("Notify", "0.7")

from gi.repository import Notify

from . import FRIENDLY_NAME, NOTIFICATION_ICON

log = logging.getLogger(__name__)


class Urgency(Enum):
    Low = 0
    Normal = 1
    Critical = 2


@dataclass
class Notification:
    summary: str
    body: Optional[str] = None
    app_icon: Optional[str] = NOTIFICATION_ICON
    urgency: Optional[Urgency] = Urgency.Normal


class NotificationServer:
    def __init__(self, app_name: str = FRIENDLY_NAME) -> None:
        log.debug("Initializing notification server.")
        Notify.init(app_name)
        self._procs: Dict[pwd.struct_passwd, Process] = {}
        self._pipes: Dict[pwd.struct_passwd, Connection] = {}
        self._update()
        log.debug("Initialized successfully.")

    @staticmethod
    def _get_active_users() -> Set[pwd.struct_passwd]:
        return set(
            pwd.getpwnam(user)
            for user in set(
                name.split()[0]
                for name in subprocess.run(["who"], capture_output=True)
                .stdout.decode()
                .splitlines()
            )
        )

    @staticmethod
    def _run_user_proc(user: pwd.struct_passwd, pipe: Connection) -> None:
        os.setgid(user.pw_gid)
        os.setuid(user.pw_uid)
        os.environ[
            "DBUS_SESSION_BUS_ADDRESS"
        ] = f"unix:path=/run/user/{user.pw_uid}/bus"
        log.debug(f"Subprocess created with uid={os.getuid()} and gid={os.getgid()}")
        try:
            temp_noti: Notify.Notification = Notify.Notification.new("No message.")
            while not pipe.closed:
                log.debug(
                    f"Blocking for new notifications from {pipe} on subprocess for user {user}"
                )
                try:
                    input = pipe.recv()
                except KeyboardInterrupt:
                    input = "QUIT"
                if isinstance(input, Notification):
                    temp_noti.update(input.summary, input.body, input.app_icon)
                    temp_noti.set_hint_byte("urgency", input.urgency.value)
                    temp_noti.show()
                elif isinstance(input, str) and input == "QUIT":
                    log.debug(f"Quit message received. Shutting down.")
                    pipe.close()
                    break
        except EOFError:
            log.debug(f"Pipe closed {pipe} on subprocess for user {user}")

    def _update(self) -> None:
        log.debug("Running subprocess update.")
        for user in self._get_active_users():
            if user not in self._procs.keys() and path.exists(
                f"/run/user/{user.pw_uid}/bus"
            ):
                # New user found, create a subprocess for them and store
                # the pipe we'll pass notifications through.
                log.debug(f"Creating new subprocess for {user=}.")
                parent_pipe, child_pipe = Pipe()
                self._procs[user] = Process(
                    target=self._run_user_proc, args=(user, child_pipe)
                )
                self._pipes[user] = parent_pipe
                self._procs[user].start()
            elif user in self._procs.keys() and not path.exists(
                f"/run/user/{user.pw_uid}/bus"
            ):
                # User must have logged out. Delete the subprocess and pipe.
                log.debug(f"Deleting stale subprocess for {user=}")
                self._pipes[user].send("QUIT")
                del self._pipes[user]
                del self._procs[user]
        log.debug("Finished subprocess update.")

    def send(self, notification: Notification) -> None:
        self._update()
        for user in self._procs.keys():
            self._pipes[user].send(notification)

    def close(self) -> None:
        for user in self._procs.keys():
            if self._procs[user].is_alive():
                try:
                    self._pipes[user].send("QUIT")
                except BrokenPipeError as e:
                    self._pipes[user].close()
                    self._procs[user].terminate()



# type: ignore
import asyncio
import logging
import os
import platform
import signal
from argparse import ArgumentParser
from os import path
from shutil import copyfile

import asuscharge
from asyncinotify import Inotify, Mask
from dbus_next.aio import MessageBus
from dbus_next.constants import BusType, RequestNameReply
from dbus_next.service import ServiceInterface, dbus_property

from . import (
    APP_NAME,
    CONFIG_DIR,
    CONFIG_PATH,
    DBUS_NAME,
    DBUS_PATH,
    STATE_PATH,
)
from .config import config
from .notify import Notification, NotificationServer

# dbus-next uses string literal type hints to determine the D-Bus type.
TUInt = "u"

log = logging.getLogger(APP_NAME)


class ChargeDaemon(ServiceInterface):
    def __init__(self) -> None:
        log.debug("Initializing daemon...")
        super().__init__(DBUS_NAME)
        log.debug(f"D-Bus service interface '{DBUS_NAME}' initialized.")
        self.controller = asuscharge.ChargeThresholdController()
        log.debug(f"Acquired ChargeThresholdController: {repr(self.controller)}")
        if config["daemon"].getboolean("notify_on_restore") or config[
            "daemon"
        ].getboolean("notify_on_change"):
            self.notifier = NotificationServer()
        else:
            self.notifier = None
        self.override_default_notification = False
        if config["daemon"].getboolean("restore_on_start"):
            try:
                threshold = int(open(STATE_PATH, "r").read().strip())
                log.debug(f"Found previous threshold state: {threshold}%.")
                if self.controller.end_threshold != threshold:
                    self.override_default_notification = True
                    self.controller.end_threshold = threshold
                    self.emit_properties_changed(
                        {"ChargeEndThreshold": self.controller.end_threshold}
                    )
                    if (
                        config["daemon"].getboolean("notify_on_restore")
                        and self.notifier
                    ):
                        log.debug("Notifying on restore.")
                        self.notifier.send(
                            Notification(
                                f"Battery charge limit set to {self.controller.end_threshold}%",
                                "Restored automatically from saved state.",
                            )
                        )
                    self.override_default_notification = False
            except FileNotFoundError:
                log.warning(
                    f"Previous threshold state not found. Using default: {self.controller.end_threshold}%."
                )
        else:
            try:
                os.remove(STATE_PATH)
                log.debug("Deleted stale threshold state file.")
            except FileNotFoundError:
                pass

    @dbus_property()
    def ChargeEndThreshold(self) -> TUInt:
        return self.controller.end_threshold

    @ChargeEndThreshold.setter
    def ChargeEndThreshold(self, value: TUInt) -> None:
        if self.controller.end_threshold == value:
            log.debug(
                f"Not updating ChargeEndThreshold: value matches current threshold."
            )
            return
        else:
            log.debug(
                f"Updating ChargeEndThreshold from {self.controller.end_threshold}% to {value}%."
            )
            self.controller.end_threshold = value
            # The shadow file is copied and the properties changed signal
            # is emitted when the threshold file itself is modified. This
            # tracks external changes too.
            # if config["daemon"].getboolean("restore_on_start"):
            #     # Shadow the charge end threshold file
            #     copyfile(self.controller.bat_path, STATE_PATH)
            # self.emit_properties_changed(
            #     {"ChargeEndThreshold": self.controller.end_threshold}
            # )


async def iterate_threshold_events(inotify: Inotify, interface: ChargeDaemon) -> None:
    async for event in inotify:
        if event.path.as_posix() == interface.controller.bat_path:
            log.debug(f"Threshold file modified, signalling property change.")
            interface.emit_properties_changed(
                {"ChargeEndThreshold": interface.controller.end_threshold}
            )
            if config["daemon"].getboolean("restore_on_start"):
                # Shadow the charge end threshold file
                copyfile(self.controller.bat_path, STATE_PATH)
            if (
                config["daemon"].getboolean("notify_on_change")
                and interface.notifier
                and not interface.override_default_notification
            ):
                interface.notifier.send(
                    Notification(
                        f"Battery charge limit updated to {interface.controller.end_threshold}%"
                    )
                )
        elif event.path.as_posix() == CONFIG_PATH:
            config.read(CONFIG_PATH)
            log.debug(
                f"Configuration file modified, reloaded values: {dict(config['daemon'].items())}"
            )


async def run_daemon() -> None:
    log.debug("Running daemon on D-Bus system bus.")
    bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
    interface = ChargeDaemon()
    log.debug(f"Exporting interface '{DBUS_NAME}' to path '{DBUS_PATH}'")
    bus.export(DBUS_PATH, interface)
    if (
        not (reply := await bus.request_name(DBUS_NAME))
        == RequestNameReply.PRIMARY_OWNER
    ):
        log.critical(
            f"Unable to acquire primary ownership of bus name '{DBUS_NAME}'. In state: {reply.name}"
        )
        raise SystemExit(1)
    log.debug("Daemon running...")
    inotify = Inotify()
    inotify.add_watch(path.dirname(interface.controller.bat_path), Mask.CLOSE_WRITE)
    inotify.add_watch(CONFIG_DIR, Mask.CLOSE_WRITE)
    tasks = asyncio.gather(
        bus.wait_for_disconnect(), iterate_threshold_events(inotify, interface)
    )

    def sig_handler(signum, frame) -> None:
        log.debug(f"Received signal: {signal.Signals(signum).name}")
        tasks.cancel()
        bus.disconnect()
        if interface.notifier:
            interface.notifier.close()
        inotify.close()
        raise SystemExit(0)

    signal.signal(signal.SIGTERM, sig_handler)
    signal.signal(signal.SIGINT, sig_handler)
    signal.signal(signal.SIGQUIT, sig_handler)
    signal.signal(signal.SIGHUP, sig_handler)


async def run_client(value: int) -> None:
    log.debug("Running client on D-Bus system bus.")
    bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
    introspection = await bus.introspect(DBUS_NAME, DBUS_PATH)
    daemon_proxy = bus.get_proxy_object(DBUS_NAME, DBUS_PATH, introspection)
    daemon_interface = daemon_proxy.get_interface(DBUS_NAME)
    log.debug(f"Acquired client interface {repr(daemon_interface)}.")
    if await daemon_interface.get_charge_end_threshold() != value:
        await daemon_interface.set_charge_end_threshold(value)
        log.debug(
            f"Set ChargeEndThreshold to {await daemon_interface.get_charge_end_threshold()}. Closing client."
        )
    else:
        log.debug(
            f"Input matches existing ChargeEndThreshold, not changing. Closing client."
        )


def main() -> None:
    if not asuscharge.supported_platform():
        raise SystemExit(
            f"{APP_NAME} only runs on Linux systems. Detected {platform.system()}."
        )
    if not asuscharge.supported_kernel():
        raise SystemExit(
            f"{APP_NAME} requires a Linux kernel version >= 5.4 to run. Detected {platform.release()}."
        )
    if not asuscharge.module_loaded():
        raise SystemExit(
            f"Required kernel module 'asus_nb_wmi' is not loaded. Try 'sudo modprobe asus_nb_wmi'."
        )
    parser = ArgumentParser(APP_NAME)
    parser.add_argument("-s", "--set", action="store", type=int)
    args = parser.parse_args()
    if args.set is not None:
        # Set the threshold with a D-Bus client and quit.
        if 1 <= args.set <= 100:
            asyncio.get_event_loop().run_until_complete(run_client(args.set))
            raise SystemExit(0)
        else:
            log.critical(f"Tried to set an invalid value: {args.set}%.")
            raise SystemExit(1)
    else:
        # Run the daemon.
        try:
            loop = asyncio.get_event_loop()
            loop.create_task(run_daemon())
            loop.run_forever()
        except Exception:
            log.exception("Daemon encountered a fatal exception.")
            raise SystemExit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asuscharged']

package_data = \
{'': ['*']}

install_requires = \
['PyGObject>=3.38.0,<4.0.0',
 'asus-charge-control>=1.0.3,<2.0.0',
 'asyncinotify>=2.0.2,<3.0.0',
 'dbus-next>=0.2.2,<0.3.0']

entry_points = \
{'console_scripts': ['asuscharged = asuscharged.__main__:main']}

setup_kwargs = {
    'name': 'asus-charge-daemon',
    'version': '0.3.1',
    'description': 'A daemon exposing D-Bus services for managing the charge level of recent ASUS notebooks.',
    'long_description': "# asus-charge-daemon\n\n> A daemon exposing D-Bus services for managing the charge level of recent\nASUS notebooks.\n\nRecent ASUS notebooks come with a Windows application to set the maximum battery\ncharge level, in order to reduce stress on the battery and prolong its lifespan. On\nLinux kernels >= version 5.4, the ```asus_nb_wmi``` kernel module exposes a sysfs object to manage this setting.\n\n```asus-charge-daemon``` is a system service that runs in the background, exposing a D-Bus interface on the System Bus, enabling userspace applications to manage the maximum battery charge level.\n\n## Installation\n\nasus-charge-daemon should work with any device running a recent kernel (>= 5.4) with the ```asus_nb_wmi``` module loaded. Use ```lsmod | grep asus_nb_wmi``` to check; if you see a line like ```asus_nb_wmi            32768  0```, then the module is running.\n\nIt has been tested with the following ASUS notebooks:\n\n- ASUS VivoBook 15 **X512DA**\n\nA Python version >= 3.7 is necessary to run this daemon. Most Linux distributions come with the right version. To verify that Python is installed on **Debian**/**Ubuntu**-based distributions, use ```apt```:\n\n```console\nsudo apt install python3\n```\n\nOn **Arch**-based distributions, use ```pacman```:\n\n```console\nsudo pacman -Syu python\n```\n\nOn **RHEL/Fedora**-based distributions, use ```rpm```:\n\n```console\nsudo rpm -i python3\n```\n\n### Manual\n\nA **very** rudimentary installation script is included for easy installation. It will copy the files to the appropriate places, install the Python modules, and run the system service.\n\nDownload and run the installation script:\n\n```console\ncurl https://raw.githubusercontent.com/cforrester1988/asus-charge-daemon/main/install.py -o install.py\nchmod +x install.py\nsudo ./install.py install\n```\n\nTo update, run the installation script again, as above. To uninstall:\n\n```console\nsudo ./install.py uninstall\n```\n\n#### git version\n\nClone the git repository locally:\n\n```console\n$ git clone https://github.com/cforrester1988/asus-charge-daemon.git\nCloning into 'asus-charge-daemon'...\n```\n\nNavigate to the directory you cloned the repository into, and run the installation script. Append ```local``` to work with the cloned package, instead of downloading it from PyPI.\n\n```console\ncd asus-charge-daemon\nsudo ./install.py install local\n```\n\nTo update, pull the latest changes and reinstall:\n\n```console\ngit pull\nsudo ./install.py reinstall local\n```\n\nTo uninstall, run the installation script again:\n\n```console\nsudo ./install.py uninstall\n```\n\n## Version history\n\n- 0.3.0 (2021-02-07)\n  - (feature) Desktop notifications\n  - (feature) Monitor threshold for outside changes\n  - (feature) Monitor config for changes\n\n- 0.2.0 (2021-01-23)\n  - Initial public release.\n",
    'author': 'Christopher Forrester',
    'author_email': 'christopher@cforrester.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com./cforrester1988/asus-charge-daemon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
